package se.umu.jobsubmission.util; 

import java.util.LinkedList; 
import java.util.List; 
import java.util.Set; 
import java.util.HashSet; 
import java.util.TreeMap; 
import java.util.Iterator; 

import EDU.oswego.cs.dl.util.concurrent.PooledExecutor;
import EDU.oswego.cs.dl.util.concurrent.PooledExecutor.BlockedExecutionHandler;
import EDU.oswego.cs.dl.util.concurrent.LinkedQueue;
import EDU.oswego.cs.dl.util.concurrent.SyncSet;
import EDU.oswego.cs.dl.util.concurrent.SyncMap;
import EDU.oswego.cs.dl.util.concurrent.Sync;
import EDU.oswego.cs.dl.util.concurrent.Mutex;
import EDU.oswego.cs.dl.util.concurrent.CondVar;
import EDU.oswego.cs.dl.util.concurrent.SynchronizedInt;

import org.apache.log4j.Logger;

public class ThreadPool {

    private static Logger logger = 
        Logger.getLogger(ThreadPool.class);

    // For synchronization between main thread and worker threads
    private Object cond; 
    private SynchronizedInt noActiveTasks; 

    private PooledExecutor pool; // manages threadpool 

    // Sets for result and additional work created during execution. 
    private SyncSet resultSet; 
    private SyncSet completedWork; 

    public ThreadPool(int noThreads) {

        logger.info("Threadpool created with " + noThreads + " threads"); 

        // Sync mutex = new Mutex();
        // this.cond = new CondVar(mutex);
        this.noActiveTasks = new SynchronizedInt(0);
        this.cond = new Object(); // used as a condition variable

        // Create pool with unbounded buffer, but no work will 
        // be stored in the pool everything that can not be executed 
        // right now is stored in workSet. 
        this.pool = new PooledExecutor(new LinkedQueue(), noThreads);
        this.pool.setKeepAliveTime(-1); // Do not kill idle threads...
        this.pool.waitWhenBlocked(); // Do not discard tasks if we are to busy. 

        // This should not be needed, but unless threads are created 
        // upon startup, they do not start to execute...
        this.pool.createThreads(noThreads); 

        // For storing results
        Set resSet = new HashSet(); 
        Sync resSync = new Mutex(); 
        this.resultSet = new SyncSet(resSet, resSync); 

        // For avoiding duplication of work
        Set completedSet = new HashSet(); 
        Sync completedSync = new Mutex(); 
        this.completedWork = new SyncSet(completedSet, completedSync); 

    }

    // Remove all old work etc. 
    // TODO: Anything else that should be reseted? 
    // --> I think not. 
    public void reset() {
        logger.info("Threadpool reseted"); 
        resultSet.clear(); 
        completedWork.clear(); 
        pool.drain(); 
        noActiveTasks.set(0); 
    }


    public void shutdown() {
        pool.shutdownNow(); 
    }

    public void execute(Runnable command) {
        try {
            pool.execute(command); 
        } catch(Exception e) {
            logger.error("Error executing " + command); 
        }
    }

    // Complete all work units and return result,  
    // start a worker for each initial work unit and wait until
    // there are no more units. 
    // Returns list of completed work. 
    public List work(List initialWorkers) {

        logger.info("Threadpool working"); 

        if(initialWorkers == null || initialWorkers.size() == 0) {
            logger.debug("No initial workers"); 
            return null; 
        }

        try {

            // Get the intial workers, and connect them with pool. 
            Worker w; 
            for(int i = 0; i < initialWorkers.size(); i++) {
                w = (Worker) initialWorkers.get(i); 
                completedWork.add(w.getWorkUnit()); 
                
                // Initial workers need to gain 
                // access to cond and noActiveTasks, workList and resultSet
                w.connectToThreadPool(resultSet, completedWork, this, 
                        noActiveTasks, cond);
                noActiveTasks.increment();   
                pool.execute(w); 
            }

            // Wait for worker threads to terminate
            synchronized (cond) {
                while(noActiveTasks.get() > 0) {
                    logger.debug("Waiting for condition"); 
                    try {
                        cond.wait(); 
                    }
                    catch (InterruptedException e) {
                        logger.error("Caught InterruptedException"); 
                    }
                }
            }
            
            logger.info("All workers done"); 

            // Always return a copy of the resultSet, 
            // ensuring that no information will be lost as we call reset()
            List result = new LinkedList(); 
            Iterator iter = resultSet.iterator(); 
            while(iter.hasNext()) 
                result.add(iter.next()); 
            
            return result; 
        } catch(InterruptedException e) {
        	logger.error("Error in threadpool execution: " + e.getMessage());            
            return null; 
        }
    }

}
