/* $Id: ngjm.h 4605 2006-01-02 20:47:04Z aleks $ */
/* ngjm.h - see ngjm.cpp for futher information */
/* Copyright 2003 Henrik Thostrup Jensen and Jesper Ryge Leth
 * All rights reserved.
 *
 * This file is part of NG Job Monitor.
 *
 * NG Job Monitor is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * NG Job Monitor is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with NGProxy; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
                                                                                          
/* 
 * NG Job Monitor is a resubmission deamon, which is a able to
 * automagically resubmit jobs that for some reason has been
 * rejected from a cluster after it was submitted.
 *
 * It pulls information about current jobs from the MDS services on
 * the cluster. If a fails, ngjm will try to figure out the
 * reason for failing and resubmit to another cluster, where the
 * should be less likely to occur (at least that is what we would
 * like it to do). In reality it just resubmits a job that has
 * failed to another cluster.
 */


#ifndef NGJM_H
#define NGJM_H
#define MAXSTATUSLENGTH	17	// PENDING:PREPARING
#define MAXGRIDJOBS	30

using namespace std;

/**
 * Function to resemble the job structure in the ng user interface 
 */
extern "C"
int ngproxyxx(const bool dryrun,
              const bool dumpxrsl,
              const int timeout,
	      const int s_timeout,
              const int debug,
              const bool anonymous);


class JobStatus {

	public:
		/**
		 * Constructer - takes xrsl string and job id
		 */
		JobStatus(Xrsl* xrsl, string jobid);

		/**
		 * Destructer
		 */
		~JobStatus();

		/**
		 * Returns the xrsl of the jobs
		 */
		Xrsl* getXrsl();

		/**
		 * Returns the job id of the job
		 */
		string getJobid();

		/**
		 * Returns the number resubmissions attempts
		 */
		int getAttempts();

		/**
		 * Returns the previous status of the job, which is one of
                 *
		 * ACCEPTED CANCELING DELETED FINISHED FINISHING INLRMS NUM
                 * PREPARING SUBMITING UNDEFINED
		 *
		 * or something...
		 */

		string getPrevStatus();

		/**
		 * Returns the time (since the epoch, in seconds) the
		 * status was last changed.
		 */

                time_t getPrevTime();

		/**
		 * Set new job status
		 */

		void setNewStatus(string s);

		/**
		 * Updates the object with new information about a submission
		 * The old cluster is put into the cluster vector, the
		 * current jobid is set to the new jobid and the number
		 * of attempts is incremented by 1
		 */
		void newSubmission(string new_jobid);

	private:
		int attempts;              // Number of submission attempts
		string current_jobid;      // Current jobid
		Xrsl* xrsl;                // Xrsl string to create object from
		vector <string> clusters;  // List of the cluster the job has been on
		string prev_status;        // Current status of job
		time_t prev_time;          // The time the status of job last changed

};


/**
 * Class to represent the resubmission daemon
 * You really shouldn't create more than one instace of this.
 */
class NGProxy {

	public:
		/**
		 * Constructer - takes the usual ui arguments +
		 * s_timeout, the timeout to wait for the job to switch
		 * state.
		 */
		NGProxy(const bool dryrun,
		        const bool dumpxrsl,
		        const int timeout,
			const int s_timeout,
		        const int debug,
		        const bool anonymous);

		/**
		 * Destructer
		 */
		~NGProxy();

		/**
		 * Starts the proxy
		 * If the boolean is true a deamon will be created
		 * If false it will just run as ordinary program (mostly
		 * for debugging. The boolean can also be omitted, since
		 * it defaults to true (create daemon).
		 * Returns 0 on succes, otherwise a negetive value is returned
		 */
		int Start(bool create_daemon);

		/**
		 * Starts to listen for connections
		 * The ONLY reason for this being public is that it can be
		 * be called from a function (uds_listener), since a thread
		 * must be created from this but we cannot take addresses
		 * of member functions, so make it public.
		 * Dont call this, unless you know what you are doing
		 */
		int Listen();

	private:
		/**
		 * Standard UI things
		 */
		bool dryrun;
		bool dumpxrsl;
		int  timeout;
		int  s_timeout;
		int  debug;
		bool anonymous;
		int  xrslsocket;
		int  maxgridjobs; // Maximun num of managed jobs
		/**
		 * Other stuff
		 */
		string ngproxy_home;    // String with path to ngproxy home
		string ngproxy_socket;  // Path to ngproxy socket
		ofstream logfile;       // Stream to logfile
		CertInfo user_proxy;    // User proxy object

		/**
		 * List of current jobs, contains objects describing the
		 * status of the jobs
		 */
		list<JobStatus*> current_jobs;

                /**
                 * List of forked download pids which have not yet been
                 * cleaned
                 */
                list<pid_t> forked_downloads;

		/**
		 * A counter indicating number of jobs in current jobs,
		 * so that we don't need to lock current jobs for long
		 * periods of time
		 */
		int num_current_jobs;
		/**
		 * corresponding lock
		 */
		pthread_mutex_t current_jobs_lock;


		/**
		 * Vector of xrsl files to be submitted
		 * The listener thread puts in a xrsl string that it recives
		 * into this, and is picked up by the mainloop which submits
		 * them. This is of course done with mutex in mind
		 * 
		 * The reason for doing it like this and not submitting the
		 * job in the listener thread is that it would require mutex
		 * on the current_jobs list, since this can be manipulated by
		 * the mainloop, while doing MDS quering, which can potentially
		 * take a lot of time. Having this queue saves us from using
		 * mutex on the jobs list (since new_jobs is only used
		 * sequentially in the mainloop). Mutex is only held for a
		 * short period on this queue.
		 */
		queue<Xrsl*> new_jobs;

		/**
		 * Lock used to ensure mutex on the new_jobs queue
		 */
		pthread_mutex_t new_jobs_lock;

		/**
		 * Prints one to three strings to the log
		 * All the given strings will be printed in one line
		 * in the log. (One call for one line).
		 */
		void Log(string msg1, string msg2, string msg3);

		/**
		 * Starts the main loop (where the stuff happens)
		 */
		void MainLoop();

		/**
		 * Handles incoming connections
		 */
		void HandleConnection(int socket_handle);

		/**
		 * Checks for arrival of new jobs
		 * And submits them accordingly
		 * 
		 * Only called from MainLoop
		 */
		void HandleNewJobs();

		/**
		 * Asks another job from the socket, if there is room in
		 * the grid. Called from HandleNewJobs
		 */
		void GetMoreJobs();
		
		/**
		 * Submits a job
		 * Takes xrsl object and string with cluster name (defaults
		 * to ""), which fill be filtered out from cluster targets.
		 * Returns an empty string on failure, otherwise it returns
		 * the jobid in a string.
		 */
		string SubmitJob(Xrsl* xrsl, string old_cluster);

		/**
		 * Pulls the status of the current jobs and if any of them
		 * are finished removed the from the current list of jobs
		 * and if any of the are failed it resbmits these jobs to
		 * other clusters
		 */
		void UpdateJobStatus();
		
		/**
		 * Handles that the job with the given jobid
		 * has failed and will attempt to handle it accordingly
		 * Returns true if the job was resubmitted
		 * False otherwise
		 */
		bool HandleJobFailure(JobStatus* js);

                /**
                 * Fetches the job from the gatekeeper so that we can
                 * actually access the stdout
                 */
                void FetchJob(const char *jobid);

};


#endif // NGHJ_H

